import collections
import logging
import os

import runez
from cached_property import cached_property

from mgit.git import GitDir, GitRunReport


__version__ = "1.2.0"
LOG = logging.getLogger(__name__)


def git_parent_path(path):
    """
    :param str path: Path to search
    :return str|None: First parent of 'path' that has a .git folder, if any
    """
    if not path or len(path) <= 5:
        return None

    if os.path.isdir(os.path.join(path, ".git")):
        return path

    return git_parent_path(os.path.dirname(path))


def find_actual_path(path):
    """
    :param str|None path: Base path, if None or '.': look for first parent folder with a .git subfolder
    :return str: Actual path to use as target
    """
    if not path or path == ".":
        path = git_parent_path(os.getcwd()) or "."

    return runez.resolved_path(path)


def get_target(path, **kwargs):
    """
    :param str path: Path to target
    :param kwargs: Optional preferences
    """
    prefs = MgitPreferences(**kwargs)
    actual_path = find_actual_path(path)
    if not actual_path or not os.path.isdir(actual_path):
        runez.abort("No folder '%s'" % runez.short(actual_path))

    if os.path.isdir(os.path.join(actual_path, ".git")):
        return GitCheckout(actual_path, prefs=prefs)

    return ProjectDir(actual_path, prefs=prefs)


def print_modified(items, color1, color2=None):
    for item in items:
        state = item[0:2]
        if color2:
            state = "%s%s" % (color1(item[0]), color2(item[1]))

        elif color1:
            state = color1(state)

        print("  %s %s" % (state, item[3:]))


class MgitPreferences:
    """Various prefs"""

    name_size = None                                # How many chars to align names when displaying list of checkouts
    align = True                                    # Whether to align names or not
    verbose = False                                 # Show verbose output
    all = False                                     # Show all entries, including missing/invalid checkout folders
    fetch = False                                   # Auto-fetch before showing status
    pull = False                                    # Auto-pull before showing status
    inspect_remotes = False                         # Inspect remote branches to report cleanable (slower)

    def __init__(self, **kwargs):
        self.update(**kwargs)

    def __repr__(self):
        result = [self._value_representation(k) for k in sorted(self.__dict__)]
        return ' '.join(s for s in result if s is not None)

    def _value_representation(self, name):
        value = getattr(self, name, None)
        if value is None:
            return None

        if value is True:
            return name

        if value is False:
            return "!%s" % name

        return "%s=%s" % (name, value)

    def set_short(self, value):
        """
        :param bool|None value: Value from parsed click command line:
                                None (default): align, verbose for individual git checkouts, short one line otherwise
                                True (--short): everything compact
                                False (--verbose): everything verbose
        """
        self.align = value is None
        self.verbose = value is False

    def update(self, **kwargs):
        for name, value in kwargs.items():
            if hasattr(self, name):
                setattr(self, name, value)
                continue

            func = getattr(self, "set_%s" % name, None)
            if func:
                func(value)
                continue

            raise Exception("Internal error: add support for flag '%s'" % name)


class RemoteProject:
    """
    Hashable object representing a remote repo
    - 'type' will indicate whether it's a stash repo, or github or other
    - 'name' will correspond to project for stash, or owner for github etc
    """

    def __init__(self, url):
        """
        :param GitURL url: URL of remote repo
        """
        self.url = url
        self.name = url.repo or "unknown"

    def __repr__(self):
        return "%s/%s" % (self.type, self.name)

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return str(self) == str(other)

    def __ne__(self, other):
        return str(self) != str(other)

    @classmethod
    def from_url(cls, url):
        if url and url.hostname:
            if "stash" in url.hostname:
                return StashProject(url)

            if "github" in url.hostname:
                return GithubProject(url)

        return UnknownProject(url)

    @property
    def type(self):
        return self.__class__.__name__.lower()[:-7]

    def projects(self):
        """
        :return dict: Projects on server hashed by their canonical name, if this is a stash server
        """
        LOG.warning("Ignoring --all option, no implementation for listing %s projects", self.type)
        return {}


class StashProject(RemoteProject):
    """Bitbucket stash repo"""


class GithubProject(RemoteProject):
    """Github repo"""


class UnknownProject(RemoteProject):
    """Unknown repo"""


class GitCheckout:
    """Represents a local git checkout"""

    def __init__(self, path, parent=None, prefs=None):
        """
        :param str path: Full path to local checkout
        :param ProjectDir|None parent: Parent project dir
        :param MgitPreferences|None prefs: Optional prefs to use
        """
        # Basename of local git folder (usually matches remote repo base name)
        self.basename = os.path.basename(path)
        self.directory_exists = os.path.isdir(path)
        self.git = GitDir(path)
        self.parent = parent
        self._prefs = prefs if prefs or parent else MgitPreferences()

    def __repr__(self):
        return self.basename

    @property
    def prefs(self):
        if self.parent:
            return self.parent.prefs

        return self._prefs

    @cached_property
    def name(self):
        """
        :return str: Basename of local git folder + remote basename if it differs
        """
        if not self.git.config.repo_name or not self.git.is_git_checkout or self.basename == self.git.config.repo_name:
            return self.basename

        return "%s (%s)" % (self.basename, self.git.config.repo_name)

    @cached_property
    def origin_project(self):
        return RemoteProject.from_url(self.git.config.origin)

    @cached_property
    def aligned_name(self):
        name = self.name
        if self.parent and self.parent.prefs.name_size:
            name = ("%%%ss" % self.parent.prefs.name_size) % name

        return name

    def header(self, report=None):
        """
        :param GitRunReport|None report: Optional report to show (defaults to self.git.report)
        :return str: Textual representation
        """
        report = GitRunReport(report or self.git.report(inspect_remotes=self.prefs.inspect_remotes))

        result = "%s:" % self.aligned_name

        if self.git.is_git_checkout:
            branch = runez.bold(self.git.branches.shortened_current_branch)
            n = len(self.git.branches.local)
            if n > 1:
                branch += " +%s" % (n - 1)

            result += " [%s]" % branch

            freshness = self.git.status.freshness
            if freshness:
                result += " %s" % freshness

        if not report.has_problems and self.parent and self.prefs and self.prefs.all and self.parent.predominant:
            if self.origin_project != self.parent.predominant:
                report.add(note="not part of %s" % self.parent.predominant)

        if report:
            rep = report.representation()
            if rep:
                result += "  %s" % rep

        return result

    def apply(self):
        """Apply switches as specified by prefs"""
        report = GitRunReport()
        if self.prefs.pull:
            if self.prefs.all and not self.git.folder_exists:
                if self.git.remote_info and self.git.remote_info.clone_url:
                    report.add(self.git.clone(self.git.remote_info.clone_url))

                else:
                    return report.cant_pull("couldn't determine clone url")

            else:
                report.add(self.git.pull())

        elif self.prefs.fetch:
            report.add(self.git.fetch())

        if not report.has_problems:
            report.add(self.git.report(inspect_remotes=self.prefs.inspect_remotes))

        return report

    def print_status(self):
        """Show checkout status"""
        report = self.apply()
        print(self.header(report))
        if self.prefs.verbose or (not self.parent and self.prefs.align):
            if len(self.git.orphan_branches) > 1:
                print("  Orphan branches: %s" % (", ".join(self.git.orphan_branches)))

            print_modified(self.git.status.modified, runez.teal, runez.red)
            print_modified(self.git.status.untracked, runez.orange)


class ProjectDir:
    """Info shown for a given directory"""

    def __init__(self, path, prefs=None):
        """
        :param str path: Path to folder
        :param MgitPreferences|None prefs: Display prefs
        """
        self.path = path                                # Path to folder to examine
        self.prefs = prefs or MgitPreferences()         # Preferences on how to output result
        self.checkouts = []                             # Actual git checkouts in 'path'
        self.projects = collections.defaultdict(set)    # Seen remotes
        self.predominant = None                         # Predominant remote, if any
        self.additional = None                          # Additional projects (sorted by checkouts, descending)
        self.stash_projects = {}                        # Corresponding projects from stash, when applicable
        self.scan()

    def __repr__(self):
        return os.path.basename(self.path)

    def scan(self):
        self.checkouts = []
        for fname in os.listdir(self.path):
            if fname and fname.startswith("."):
                continue

            source_path = os.path.join(self.path, fname)
            if os.path.isdir(source_path):
                r = GitCheckout(source_path, parent=self)
                self.checkouts.append(r)
                if r.git.is_git_checkout:
                    self.projects[r.origin_project].add(r)

        self.predominant = None
        self.additional = None
        counts = [(project, len(self.projects[project])) for project in sorted(self.projects, key=lambda x: -len(self.projects[x]))]
        if counts:
            self.additional = [t[0] for t in counts]
            top, top_count = counts.pop(0)
            threshold = top_count // 2
            if not counts or all(t[1] <= threshold for t in counts):
                self.predominant = top
                self.additional = self.additional[1:]

        if not self.prefs.all or not self.predominant:
            self.stash_projects = {}

        else:
            self.stash_projects = self.predominant.projects()
            seen = {}
            for checkout in self.checkouts:
                if not checkout.git.is_git_checkout:
                    continue

                canonical_name = checkout.git.config.repo_name
                seen[canonical_name] = True
                checkout.git.remote_info = self.stash_projects.get(canonical_name)

            for name, project in self.stash_projects.items():
                if name in seen:
                    continue

                path = os.path.join(self.path, name)
                if os.path.isdir(path):
                    path += ".1"

                r = GitCheckout(path, parent=self)
                r.git.remote_info = project
                self.checkouts.append(r)

        self.checkouts = sorted(self.checkouts, key=lambda x: x.basename)
        if self.prefs.align and self.projects:
            self.prefs.name_size = min(36, max(len(c.name) for c in self.checkouts))

        else:
            self.prefs.name_size = None

    @cached_property
    def header(self):
        result = "%s:" % runez.purple(runez.short(self.path))

        if not self.projects:
            return "%s %s" % (result, runez.orange("no git folders"))

        if self.predominant:
            result += runez.bold(" %s %s" % (len(self.projects[self.predominant]), self.predominant))

        else:
            result += runez.orange(" no predominant project")

        if self.additional:
            result += " (%s)" % runez.purple(", ".join("+%s %s" % (len(self.projects[project]), project) for project in self.additional))

        return result

    def print_status(self):
        """Show checkout status"""
        print(self.header)
        for checkout in self.checkouts:
            if self.prefs.all or checkout.git.is_git_checkout:
                checkout.print_status()



import collections
import logging
import os
import re
import subprocess  # nosec
import time

try:
    from urlparse import urlparse

except ImportError:
    from urllib.parse import urlparse

import runez
from cached_property import cached_property


LOG = logging.getLogger(__name__)
FETCH_AGE_FILES = ["FETCH_HEAD", "HEAD"]
FRESHNESS_THRESHOLD = 12 * runez.date.SECONDS_IN_ONE_HOUR
BRANCH_INVALID_CHARS = "~^: \t\\"
GIT_ERROR_PREFIXES = {"git", "error", "fatal"}

RE_GITHUB_SSH = re.compile(r"^git@([^:]+):(\w+)/([^/]+)$")
RE_BRANCH_STATUS = re.compile(r"^## (.+)\.\.\.(([^/]+)/)?([^ ]+)\s*(\[(.+)])?$")


def reset_cached_properties(obj, names=None):
    """
    :param obj: Object which cached properties to reset
    :param str|None names: Optional list of cached property names to reset (all properties if no 'names' given)
    """
    if not obj:
        return

    if names:
        if hasattr(names, "startswith") and names.startswith("-"):
            names = set(obj.__class__.__dict__) - set(names[1:].split())

        elif hasattr(names, "split"):
            names = names.split()

    else:
        names = obj.__class__.__dict__

    for k in names:
        v = getattr(obj.__class__, k, None)
        if isinstance(v, cached_property) and k in obj.__dict__:
            del obj.__dict__[k]


def is_valid_branch_name(name):
    """
    :param str|None name: Branch name to validate
    :return bool: True if branch name appears valid, as per https://wincent.com/wiki/Legal_Git_branch_names
    """
    if not name or name[0] == "." or ".." in name or name.endswith("/") or name.endswith(".lock"):
        return False

    for char in name:
        if ord(char) < 32 or char in BRANCH_INVALID_CHARS:
            return False

    return True


def shortened_message(text, keep_lines=2, separator=" "):
    """
    :param str text: Original git error message (those can be verbose, and include progress)
    :param int keep_lines: Max lines to keep
    :param str separator: Lines are split for shortening, separator to use to re-join lines
    :return str: Shortened git error message
    """
    lines = []
    prefixed = []
    for line in text.strip().split("\n"):
        line = line.strip().strip(".")
        if not line:
            continue

        p = line.partition(":")
        if p[2] and p[0] in GIT_ERROR_PREFIXES:
            prefixed.append(p[2].strip())

        else:
            lines.append(line)

    if prefixed:
        lines = prefixed

    if keep_lines and len(lines) > keep_lines:
        lines = lines[:keep_lines]

    return separator.join(lines).replace("  ", " ").strip()


class GitRunReport:
    """Convenient and easy to compose reporting class"""

    def __init__(self, *args, **kwargs):
        self._progress = []
        self._note = []
        self._problem = []
        self.add(*args, **kwargs)

    def __repr__(self):
        return "%s problems, %s progress, %s notes" % (len(self._problem), len(self._progress), len(self._note))

    def __contains__(self, text):
        """
        :param str text: Text to look up
        :return bool: True if 'text' is mentioned in one of the messages in self._problem
        """
        if not text:
            return False

        for problem in self._problem:
            if text in problem:
                return True

        return False

    @classmethod
    def not_git(cls):
        return GitRunReport(problem="<not a git checkout")

    def cant_pull(self, reason=None):
        self.add(problem="<can't pull")
        if reason:
            self.add(problem=reason)

        return self

    @property
    def has_problems(self):
        return bool(self._problem)

    def representation(self, progress=True, note=True, max_chars=160, separator="; "):
        """
        :param bool progress: Show repos with progress mention (pulled/cloned)
        :param bool note: Show repos with notes
        :param int max_chars: Max chars to show (truncate if messages are longer)
        :param str separator: Separator to use
        :return str: Textual representation
        """
        result = []
        n = _add_sorted(result, self._problem, runez.red, 0, max_chars)

        if progress:
            n = _add_sorted(result, self._progress, runez.plain, n, max_chars)

        if note:
            _add_sorted(result, self._note, runez.purple, n, max_chars)

        result = separator.join(result)
        if len(result) > max_chars:
            result = "%s..." % (result[: max_chars - 3])

        return result

    def _add(self, target, items):
        """
        :param list target: Where to add 'items'
        :param items: items to add
        """
        if not items:
            return

        if isinstance(items, (list, tuple)):
            for item in items:
                self._add(target, item)

        elif items not in target:
            target.append(items)

    def cumulate(self, other):
        """
        :param GitRunReport other: Cumulate 'other' with current report
        :return GitRunReport: Returns self
        """
        if isinstance(other, GitRunReport):
            self._add(self._progress, other._progress)
            self._add(self._note, other._note)
            self._add(self._problem, other._problem)

        return self

    def add(self, *args, **kwargs):
        """
        :param args: Optional, other reports to cumulate
        :param kwargs: Optional, attributes to add
        :return GitRunReport: Returns self
        """
        for item in args:
            self.cumulate(item)

        for key, value in kwargs.items():
            attribute_name = "_%s" % key
            target = getattr(self, attribute_name, None)
            if target is None:
                raise Exception("Internal error: invalid GitRunReport target '%s'" % key)

            if isinstance(value, (list, tuple)):
                for item in value:
                    self.add(**{key: item})

            elif isinstance(value, GitRunReport):
                self._add(target, getattr(value, attribute_name))

            else:
                self._add(target, value)

        return self


class GitURL:
    """Parse and extract meaningful info from a git repo url"""

    def __init__(self):
        self.url = None
        self.protocol = None
        self.hostname = None
        self.relative_path = None
        self.username = None
        self.name = None
        self.repo = None

    def __repr__(self):
        return self.url or ""

    def _set_name(self, basename):
        """
        :param str|None basename: Set 'self.name' from 'basename' of url
        """
        if basename and basename.endswith(".git"):
            basename = basename[:-4]

        self.name = basename or "unknown"

    def _set_repo(self, dirname):
        """
        :param str|None dirname: Set 'self.repo' from 'dirname' of url
        """
        if dirname and "/" in dirname:
            dirname = os.path.basename(dirname)

        self.repo = dirname or "unknown"

    def set(self, url):
        """
        :param str url: Set fields of this object, extracted from git repo 'url'
        """
        self.url = url or ""
        if not url:
            self.protocol = "unknown"
            self.hostname = "unknown"
            self.relative_path = ""
            self.username = None
            self._set_name(None)
            self._set_repo(None)
            return

        if url.startswith("git@"):
            m = RE_GITHUB_SSH.match(url)
            if m:
                self.protocol = "ssh"
                self.hostname = m.group(1) or "unknown"
                self.relative_path = "%s/%s" % (m.group(2), m.group(3))
                self.username = "git"
                self._set_name(m.group(3))
                self._set_repo(m.group(2))
                return
            url = "ssh://%s" % url

        p = urlparse(url)
        self.protocol = p.scheme or "file"
        self.hostname = p.hostname or "local"
        self.relative_path = p.path.rstrip("/")
        self.username = p.username
        self._set_name(os.path.basename(self.relative_path))
        self._set_repo(os.path.dirname(self.relative_path))


class GitDir:
    """Model a local git repo"""

    def __init__(self, path):
        """
        :param str path: Path to local repo
        """
        self.path = path
        self.folder_exists = os.path.exists(path)
        self.is_git_checkout = self.folder_exists and os.path.isdir(os.path.join(path, ".git"))
        self.remote_info = None

    def __repr__(self):
        if not self.is_git_checkout:
            return "! %s" % self.path

        return self.path

    def report(self, bare=False, inspect_remotes=False):
        """
        :param bool bare: Bare report only
        :param bool inspect_remotes: If True, report on which remote branches are cleanable
        :return GitRunReport: General report on current checkout state
        """
        if not self.is_git_checkout:
            if self.remote_info:
                return GitRunReport(problem="not cloned yet")

            return GitRunReport.not_git()

        result = GitRunReport()

        if not self.config.remotes:
            result.add(problem="no remotes")

        if bare:
            return result

        if self.age is not None and self.age > FRESHNESS_THRESHOLD:
            result.add(note="last fetch %s ago" % runez.represented_duration(self.age))

        orphan_branches = self.orphan_branches
        if self.branches.current in orphan_branches:
            # Current is no more on its remote (should possibly checkout another branch and cleanup, or push)
            orphan_branches = orphan_branches[:]
            orphan_branches.remove(self.branches.current)
            result.add(note="current branch '%s' is orphaned" % self.branches.current)

        if len(orphan_branches) == 1:
            result.add(note="local branch '%s' can be pruned" % orphan_branches[0])

        elif orphan_branches:
            result.add(note="%s can be pruned" % runez.plural(orphan_branches, "local branch"))

        result.add(self.branches.report)

        if inspect_remotes and self.remote_cleanable_branches:
            if len(self.remote_cleanable_branches) == 1:
                cleanable = "'%s'" % next(iter(self.remote_cleanable_branches))

            else:
                cleanable = runez.plural(self.remote_cleanable_branches, "remote branch")

            result.add(note="%s can be cleaned" % cleanable)

        return result

    def _git_command(self, args):
        """
        :param list|tuple args: Git command + args to use
        :return list, str: Full git invocation + human friendly representation
        """
        cmd = ["git"]
        if args and args[0] == "clone":
            args_represented = "git %s" % " ".join(args)

        else:
            args_represented = "git -C %s %s" % (runez.short(self.path), " ".join(args))
            cmd.extend(["-C", self.path])

        cmd.extend(args)
        return cmd, args_represented

    def run_raw_git_command(self, *args):
        """
        :param args: Execute git command with provided args, don't capture its output, but let it show through stdout/stderr
        :return GitRunReport: Report
        """
        cmd, pretty_args = self._git_command(args)
        pretty_args = "git %s" % " ".join(args)
        print("Running: %s" % runez.bold(pretty_args))
        proc = subprocess.Popen(cmd)  # nosec
        proc.communicate()
        if proc.returncode:
            return GitRunReport(problem="git exited with code %s" % proc.returncode)

        return GitRunReport()

    def run_git_command(self, *args):
        """
        :param args: Execute git command with provided args
        :return str, GitRunReport: Output from git command + report on eventual error
        """
        cmd, pretty_args = self._git_command(args)
        LOG.debug("Running: %s", pretty_args)
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)  # nosec
        output, error = proc.communicate()
        output = runez.decode(output)
        error = runez.decode(error)
        if proc.returncode == 0:
            return output, GitRunReport()

        if not error:
            return output, GitRunReport(problem="git exited with code %s" % proc.returncode)

        return output, GitRunReport(problem=shortened_message(error))

    def fallback_branch(self):
        """
        :return str: Best branch to fallback to if we need to clean or reset
        """
        if self.branches.current not in self.orphan_branches:
            return self.branches.current

        for branch in self.branches.local:
            if branch not in self.orphan_branches:
                return branch

        return self.branches.default_branches.get("origin")

    def reset_cached_properties(self, partial=False):
        """Reset cached properties that may have changed after a fetch or pull"""
        if partial:
            reset_cached_properties(self, "-config")

        else:
            reset_cached_properties(self)

    def fetch(self, age=30):
        """
        :param int|None age: Fetch if age is older than specified number of seconds, use None to fetch unconditionally
        :return GitRunReport:
        """
        if not self.is_git_checkout:
            return GitRunReport.not_git()

        if age is not None:
            current_age = self.age
            if current_age is not None and current_age <= age:
                return GitRunReport()

        _, error = self.run_git_command("fetch", "--all", "--prune")
        self.reset_cached_properties(partial=True)
        return error

    def pull(self):
        """Pull from tracked remote"""
        if not self.is_git_checkout:
            return GitRunReport.not_git().cant_pull()

        report = self.report(bare=True)
        if report.has_problems:
            return report.cant_pull()

        if self.status.modified:
            return GitRunReport().cant_pull("pending changes")

        if self.status.report.has_problems:
            return GitRunReport(problem=self.status.report).cant_pull()

        if not self.branches.current:
            return GitRunReport(problem="no remote branch")

        if self.branches.current == "HEAD" and self.branches.current in self.orphan_branches:
            # Untracked HEAD
            branch = self.fallback_branch()
            if not branch:
                return GitRunReport(problem="can't determine fallback branch")

            output, error = self.run_git_command("checkout", branch)
            if error.has_problems:
                self.reset_cached_properties(partial=True)
                return error

        output, error = self.run_git_command("pull", "--rebase")
        self.reset_cached_properties(partial=True)

        if error.has_problems:
            if "following untracked" in error:
                return GitRunReport().cant_pull("untracked files would be overwritten")

            if "Repository not found" in error:
                return GitRunReport().cant_pull("repository not found")

            return error.cant_pull()

        if "up to date" in output or "up-to-date" in output:
            return GitRunReport(progress="")

        if "Fast-forward" in output:
            return GitRunReport(progress="pulled successfully")

        # Shouldn't be reached
        LOG.debug("Check pull --rebase output: %s, error: %s", output, error)
        lines = []
        if output:
            lines.extend(s.strip() for s in output.strip().split("\n") if s.strip())

        lines.append(error.representation(progress=False, note=False).strip())
        output = lines[0] if lines else "no output"
        return GitRunReport(note="pull may have been unsuccessful (%s)" % output)

    def clone(self, url):
        if self.folder_exists:
            return GitRunReport(problem="folder already exists, can't clone")

        output, error = self.run_git_command("clone", url, self.path)
        self.folder_exists = os.path.exists(self.path)
        self.is_git_checkout = self.folder_exists and os.path.isdir(os.path.join(self.path, ".git"))
        self.reset_cached_properties()

        if error.has_problems:
            return error.add(problem="<can't clone")

        return GitRunReport(progress="cloned successfully")

    @cached_property
    def age(self):
        """
        :return int|None: Elapsed time in seconds since last fetch
        """
        for name in FETCH_AGE_FILES:
            try:
                last_fetch = os.path.getmtime(os.path.join(os.path.join(self.path, ".git"), name))
                return int(time.time() - last_fetch)

            except OSError:
                pass

    @cached_property
    def status(self):
        """
        :return GitStatus: Parsed info from 'git status --porcelain --branch'
        """
        return GitStatus(self)

    @cached_property
    def config(self):
        """
        :return GitConfig: Parsed info from 'git config --list'
        """
        return GitConfig(self)

    @cached_property
    def branches(self):
        """
        :return GitConfig: Parsed info from 'git branch --list --all'
        """
        return GitBranches(self)

    @cached_property
    def orphan_branches(self):
        """
        :return list(str): Local branch names that were deleted on their corresponding remote
        """
        result = []
        for name in self.branches.local:
            remote = self.config.tracking_remote.get(name)
            if not remote or remote not in self.branches.by_remote or name not in self.branches.by_remote[remote]:
                result.append(name)

        return result

    @cached_property
    def special_branches(self):
        result = set(self.branches.default_branches.values())
        result.add("HEAD")
        result.add("master")
        result.add("test")
        result.add("prod")
        return result

    @cached_property
    def local_cleanable_branches(self):
        """
        :return set: Local branches that can be cleaned
        """
        result = set(name for name in self.orphan_branches if name not in self.special_branches)
        for branch in self.remote_cleanable_branches:
            remote, _, name = branch.partition("/")
            tracking = self.config.tracking_remote.get(name)
            if tracking == remote:
                result.add(name)

        return result

    @cached_property
    def remote_cleanable_branches(self):
        """
        :return set: Remote branches that can be cleaned
        """
        result = set()
        aspect = GitBranches(self, auto_load=False)
        aspect._command = "branch --list --remote --merged"
        aspect._remote_prefix = ""
        default = self.branches.default_branches.get("origin")
        if default:
            aspect._command += " %s" % default

        aspect.reload()
        for remote, branches in aspect.by_remote.items():
            url = self.config.remotes.get(remote)
            if not url or url.protocol != "ssh":
                continue

            result.update(["%s/%s" % (remote, branch) for branch in branches if branch not in self.special_branches])

        return result


class GitAspect:
    """Common ancestor for info gathered from git"""

    _command = None

    def __init__(self, parent, auto_load=True):
        self._parent = parent
        self._lines = None  # Lines from output of last command run, for troubleshooting
        if auto_load:
            self.reload()

    def __repr__(self):
        return self._command

    def reload(self):
        for k in self.__class__.__dict__:
            if k.startswith("_"):
                continue

            v = getattr(self.__class__, k, None)
            if v is None or isinstance(v, (property, cached_property)) or callable(v):
                continue

            if isinstance(v, collections.defaultdict):
                v = collections.defaultdict(v.default_factory)

            else:
                v = v.__class__()

            setattr(self, k, v)

        if not self._parent.is_git_checkout:
            return

        output, error = self._parent.run_git_command(*self._command.split())
        if error.has_problems:
            LOG.debug("Prev git command had error output: [%s]", error.representation())

        self._lines = [line for line in output.split("\n") if line.strip()]
        for line in self._lines:
            self._process_line(line)

    def _process_line(self, line):
        raise Exception("Not implemented")


class GitBranches(GitAspect):
    """Branch info"""

    _command = 'branch --list --all'
    _remote_prefix = 'remotes/'

    current = ''                                    # Current local branch
    local = set()                                   # Local branches
    by_remote = collections.defaultdict(set)        # Branches by remote (usually origin and optionally upstream)
    default_branches = {}                           # Default branch per remote
    report = GitRunReport()

    @property
    def shortened_current_branch(self):
        return str(self.current or 'HEAD').replace('feature/', 'f/').replace('bugfix/', 'b/')

    def _process_line(self, line):
        if not line or len(line) <= 3 or line[0] not in ' *' or line[1] != ' ':
            LOG.warning("Internal error: malformed branch --list line: %s", line)
            return

        name = line[2:]
        if name.startswith(self._remote_prefix):
            name = name[len(self._remote_prefix):]
            default = None
            try:
                i = name.index(' -> ')
                first = name[:i]
                if first.endswith('/HEAD'):
                    default = name = name[i + 4:]

            except ValueError:
                pass

            remote, _, name = name.partition('/')
            self.by_remote[remote].add(name)
            if default:
                self.default_branches[remote] = name

            return

        if name.startswith('('):
            name = name[1:]
            if name.endswith(')'):
                name = name[:-1]

            name, _, problem = name.partition(' ')
            self.report.add(note='%s %s' % (name, problem))

        self.local.add(name)
        if line[0] == '*':
            self.current = name


class GitConfig(GitAspect):
    """Remote info"""

    _command = 'config --list'
    origin = GitURL()                           # URL to remote called 'origin'
    remotes = {}                                # GitURL by remote name map
    tracking_remote = {}                        # Remotes that each local branch is tracking
    content = {}

    @cached_property
    def repo_name(self):
        """
        :return str: Most significant repository name
        """
        if self.origin:
            return self.origin.name

        for r in self.remotes.values():
            return r.name

        return None

    def _process_line(self, line):
        k, _, v = line.partition("=")
        self.content[k] = v
        if k.startswith("remote."):
            if k.endswith(".url"):
                k = k[7:-4]
                url = GitURL()
                url.set(v)
                self.remotes[k] = url
                if k == "origin":
                    self.origin = url

        elif k.startswith("branch."):
            if k.endswith(".remote"):
                self.tracking_remote[k[7:-7]] = v


class GitStatus(GitAspect):
    """Currently modified files"""

    _command = "status --porcelain --branch"

    modified = []
    untracked = []
    report = GitRunReport()

    @property
    def freshness(self):
        """Short freshness overview"""
        result = []
        if self.report._problem:
            result.append(runez.red(" ".join(self.report._problem)))

        if self.modified:
            result.append(runez.red(runez.plural(self.modified, "diff")))

        if self.untracked:
            result.append(runez.orange("%s untracked" % len(self.untracked)))

        if self.report._note:
            result.append(runez.purple(" ".join(self.report._note)))

        if not self.report._problem and not self.report._note and self._parent.age is not None:
            message = "up to date"
            if self._parent.age > FRESHNESS_THRESHOLD:
                message += "*"

            result.append(runez.teal(message))

        return ", ".join(result)

    def _process_line(self, line):
        if line[0] == "#":
            if "..." not in line:
                return

            m = RE_BRANCH_STATUS.match(line)
            if not m:
                LOG.warning("Unrecognised git status line: '%s'", line)
                return

            text = str(m.group(6) or "")  # behind, ahead, or gone
            if not text:
                return

            for message in text.split(","):
                message = message.strip()
                if "gone" in message:
                    line = line.lower()
                    if "no commits yet" in line or "initial commit on" in line:
                        self.report.add(note="no commits yet")

                    else:
                        self.report.add(problem="remote branch gone")

                elif "ahead" in message:
                    self.report.add(problem=message)

                else:
                    self.report.add(note=message)

            return

        if line[0] == "?":
            self.untracked.append(line)
            return

        self.modified.append(line)


def _report_sorter(enum):
    """
    :param tuple(int, str) enum: Tuple from enumerate()
    :return int: Value to use for sorting messages in this report
    """
    index, message = enum
    if message[0] == '<':
        return -enum[0]                 # '<' makes message sort towards front, but keeping order with other such prefixed messages

    if message[0] == '>':
        return 1000000 + enum[0]        # '>' makes message sort towards end

    return enum[0]                  # Non-prefixed message stay where they were


def _add_sorted(result, target, color, n, max_chars):
    """
    :param list(str) result: Where to accumulate sorted report
    :param list(str) target: Target to sort, respecting '<' and '>' prefixing
    :param color: Optional color to use
    :param int n: How many chars were consumed so far
    :param int|None max_chars: Maximum number of characters to yield
    :return int: Number of chars accumulated
    """
    if max_chars and n > max_chars:
        # We already reached limit
        return n

    items = []
    for message in (s.lstrip("<>") for i, s in sorted(enumerate(target), key=_report_sorter)):
        size = len(message)
        if max_chars:
            remaining = max_chars - n
            if remaining < size:
                items.append(message[:remaining])
                n += size
                break

        n += size
        items.append(message)

    result.extend(color(s) for s in items)
    return n



from setuptools import setup


setup(
    name="mgit",
    setup_requires="setupmeta",
    versioning="dev",
    author="Zoran Simic zoran@simicweb.com",
    keywords='multiple, git, repos',
    url="https://github.com/zsimic/mgit",
    entry_points={
        "console_scripts": [
            "mgit = mgit.cli:main",
        ],
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Software Development :: Build Tools",
        "Topic :: Utilities"
    ],
)



"""
\b
Advanced usage:
  --clean show                  Show which local/remote branches can be cleaned
  --clean local                 Clean local branches that were deleted from their corresponding remote
  --clean remote                Clean merged remote branches
  --clean all                   Clean local and merged remote branches
  --clean reset                 Do a git --reset --hard + clean -fdx (nuke all changes, get back to pristine state)
\b
"""

import logging
import sys

import click
import runez

from mgit import get_target, GitCheckout
from mgit.git import GitRunReport


LOG = logging.getLogger(__name__)
VALID_CLEAN_ACTIONS = "show local remote all reset".split()


@runez.click.command()
@runez.click.version()
@runez.click.debug()
@runez.click.color()
@runez.click.log()
@click.option("--clean", default=None, type=click.Choice(VALID_CLEAN_ACTIONS), help="Auto-clean branches")
@click.option("-f", "--fetch", is_flag=True, default=False, help="Fetch from all remotes")
@click.option("-p", "--pull", is_flag=True, default=False, help="Pull from tracking remote")
@click.option("-s/-v", "--short/--verbose", is_flag=True, default=None, help="Short/verbose output")
@click.option("-cs", is_flag=True, default=False, help="Handy shortcut for '--clean show'")
@click.option("-cl", is_flag=True, default=False, help="Handy shortcut for '--clean local'")
@click.option("-cr", is_flag=True, default=False, help="Handy shortcut for '--clean remote'")
@click.option("-ca", is_flag=True, default=False, help="Handy shortcut for '--clean all'")
@click.argument("target", required=False, default=None)
def main(debug, log, clean, target, **kwargs):
    """
    Fetch collections of git projects
    """
    runez.system.AbortException = SystemExit
    runez.date.DEFAULT_DURATION_SPAN = -2
    runez.log.setup(debug=debug, console_format="%(levelname)s %(message)s", file_location=log, locations=None)

    hc = handy_clean(kwargs)
    if not clean:
        clean = hc

    target = get_target(target, **kwargs)

    if clean is not None:
        handle_clean(target, clean)
        sys.exit(0)

    target.print_status()


def handy_clean(kwargs):
    """
    :param kwargs: Pop all handy shortcuts from kwargs
    :return str|None: Equivalent full --clean option
    """
    cs = kwargs.pop("cs")
    cl = kwargs.pop("cl")
    cr = kwargs.pop("cr")
    ca = kwargs.pop("ca")
    if cs:
        return "show"

    if cl:
        return "local"

    if cr:
        return "remote"

    if ca:
        return "all"

    return None


def run_git(target, fatal, *args):
    """Run git command on target, abort if command exits with error code"""
    error = target.git.run_raw_git_command(*args)
    if error.has_problems:
        if fatal:
            runez.abort(error.representation())

        print(error.representation())
        return 0

    return 1


def clean_reset(target):
    """
    :param GitCheckout target: Target to reset
    """
    fallback = target.git.fallback_branch()
    if not fallback:
        runez.abort("Can't determine a branch that can be used for reset")

    run_git(target, True, "reset", "--hard", "HEAD")
    run_git(target, True, "clean", "-fdx")
    if fallback != target.git.branches.current:
        run_git(target, True, "checkout", fallback)

    run_git(target, True, "pull")
    target.git.reset_cached_properties()
    print(target.header())


def clean_show(target):
    """
    :param GitCheckout target: Target to show
    """
    print(target.header())
    if not target.git.local_cleanable_branches:
        print("  No local branches can be cleaned")

    else:
        for branch in target.git.local_cleanable_branches:
            print("  %s branch %s can be cleaned" % (runez.bold("local"), runez.bold(branch)))

    if not target.git.remote_cleanable_branches:
        print("  No remote branches can be cleaned")

    else:
        for branch in target.git.remote_cleanable_branches:
            print("  %s can be cleaned" % (runez.bold(branch)))


def handle_single_clean(target, what):
    """
    :param GitCheckout target: Single checkout to clean
    :param str what: Operation
    """
    report = target.git.fetch()
    if report.has_problems:
        if what != "reset":
            what = "clean"

        print(target.header(GitRunReport(report).add(problem="<can't %s" % what)))
        runez.abort("")

    if what == "reset":
        return clean_reset(target)

    if what == "show":
        return clean_show(target)

    total_cleaned = 0
    print(target.header())

    if what in "remote all":
        if not target.git.remote_cleanable_branches:
            print("  No remote branches can be cleaned")

        else:
            total = len(target.git.remote_cleanable_branches)
            cleaned = 0
            for branch in target.git.remote_cleanable_branches:
                remote, _, name = branch.partition("/")
                if not remote and name:
                    raise Exception("Unknown branch spec '%s'" % branch)

                if run_git(target, False, "branch", "--delete", "--remotes", branch):
                    cleaned += run_git(target, False, "push", "--delete", remote, name)

            total_cleaned += cleaned
            if cleaned == total:
                print("%s cleaned" % runez.plural(cleaned, "remote branch"))

            else:
                print("%s/%s remote branches cleaned" % (cleaned, total))

            target.git.reset_cached_properties()
            if what == "all":
                # Fetch to update remote branches (and correctly detect new dangling local)
                target.git.fetch()

    if what in "local all":
        if not target.git.local_cleanable_branches:
            print("  No local branches can be cleaned")

        else:
            total = len(target.git.local_cleanable_branches)
            cleaned = 0
            for branch in target.git.local_cleanable_branches:
                if branch == target.git.branches.current:
                    fallback = target.git.fallback_branch()
                    if not fallback:
                        print("Skipping branch '%s', can't determine fallback branch" % target.git.branches.current)
                        continue

                    run_git(target, True, "checkout", fallback)
                    run_git(target, True, "pull")

                cleaned += run_git(target, False, "branch", "--delete", branch)

            total_cleaned += cleaned
            if cleaned == total:
                print(runez.bold("%s cleaned" % runez.plural(cleaned, "local branch")))

            else:
                print(runez.orange("%s/%s local branches cleaned" % (cleaned, total)))

            target.git.reset_cached_properties()

    if total_cleaned:
        print(target.header())


def handle_clean(target, what):
    if isinstance(target, GitCheckout):
        handle_single_clean(target, what)
        return

    if what in "remote reset":
        runez.abort("Only '--clean show' and '--clean local' supported for multiple git checkouts for now")

    target.prefs.name_size = None
    target.prefs.set_short(True)
    for sub_target in target.checkouts:
        handle_single_clean(sub_target, what)
        print("----")
